"""
Script to move a subset of GridFS files from one db
(or collection) to another.

>>> import io
>>> db_uri = getfixture('mongodb_uri') + '/move_gridfs_test'
>>> source = helper.connect_gridfs(db_uri + '.source')
>>> dest = helper.connect_gridfs(db_uri + '.dest')
>>> id = source.put(io.BytesIO(b'test'), filename='test.txt')
>>> mover = FileMove(source_gfs=source, dest_gfs=dest, delete=True)
>>> mover.ensure_indexes()
>>> mover.run(bar=None)
>>> source.list()
[]
>>> dest.exists(id)
True
>>> dest.list()
['test.txt']
"""

import sys
import logging
import argparse
import itertools
import signal

from six.moves import map

import bson
import dateutil.parser
from jaraco.ui import progress
from more_itertools.recipes import consume

from jaraco.mongodb import helper


log = logging.getLogger()


def get_args():
	parser = argparse.ArgumentParser()
	parser.add_argument('source_gfs', type=helper.connect_gridfs)
	parser.add_argument('dest_gfs', type=helper.connect_gridfs)
	parser.add_argument(
		'--include',
		help="a filter of files (regex) to include",
	)
	parser.add_argument(
		'--delete', default=False, action="store_true",
		help="delete files after moving",
	)
	parser.add_argument('--limit', type=int)
	parser.add_argument(
		'--limit-date', type=dateutil.parser.parse,
		help="only move files older than this date",
	)
	return parser.parse_args()


class FileMove:
	include = None
	delete = False
	limit = None
	limit_date = None

	def __init__(self, **params):
		vars(self).update(**params)

	def ensure_indexes(self):
		"""
		Create the same indexes that the GridFS API would have
		"""
		self.dest_gfs.new_file()._GridIn__ensure_indexes()

	@property
	def filter(self):
		filter = {}
		if self.include:
			filter.update(filename={"$regex": self.include})
		if self.limit_date:
			id_max = bson.objectid.ObjectId.from_datetime(self.limit_date)
			filter.update(_id={"$lt": id_max})
		return filter

	@property
	def source_coll(self):
		return self.source_gfs._GridFS__collection

	@property
	def dest_coll(self):
		return self.dest_gfs._GridFS__collection

	def run(self, bar=progress.TargetProgressBar):
		files = self.source_coll.files.find(
			self.filter,
			batch_size=1,
		)
		limit_files = itertools.islice(files, self.limit)
		count = min(files.count(), self.limit or float('inf'))
		progress = bar(count).iterate if bar else iter
		with SignalTrap(progress(limit_files)) as items:
			consume(map(self.process, items))

	def process(self, file):
		chunks = self.source_coll.chunks.find(dict(files_id=file['_id']))
		for chunk in chunks:
			self.dest_coll.chunks.insert(chunk)
		self.dest_coll.files.insert(file)
		self.delete and self.source_gfs.delete(file['_id'])


class SignalTrap:
	"""
	A context manager for wrapping an iterable such that it
	is only interrupted between iterations.
	"""

	def __init__(self, iterable):
		self.iterable = iterable

	def __enter__(self):
		self.prev = signal.signal(signal.SIGINT, self.stop)
		return self

	def __exit__(self, *args):
		signal.signal(signal.SIGINT, self.prev)
		del self.prev

	def __iter__(self):
		return self

	def __next__(self):
		return next(self.iterable)
	next = __next__

	def stop(self, signal, frame):
		self.iterable = iter([])


def run():
	logging.basicConfig(stream=sys.stderr, level=logging.INFO)
	args = get_args()

	mover = FileMove(**vars(args))
	mover.ensure_indexes()
	mover.run()


if __name__ == '__main__':
	run()



import shlex
import contextlib

import pmxbot.storage

from . import sharding


class Storage(pmxbot.storage.SelectableStorage, pmxbot.storage.MongoDBStorage):
	collection_name = 'unused'


def get_client():
	"""
	Use the same MongoDB client as pmxbot if available.
	"""
	with contextlib.suppress(Exception):
		store = Storage.from_URI()
		assert isinstance(store, pmxbot.storage.MongoDBStorage)
		return store.db.database.client


@pmxbot.core.command("create-db-in-shard")
def cdbs(client, event, channel, nick, rest):
	"""
	Create a database in a shard. !create-db-in-shard {db name} {shard name}
	"""
	db_name, shard = shlex.split(rest)
	return sharding.create_db_in_shard(db_name, shard, client=get_client())



from jaraco.mongodb import compat


def test_save_no_id(database):
	doc = dict(foo='bar')
	compat.save(database.test_coll, doc)
	assert database.test_coll.find_one()['foo'] == 'bar'


def test_save_new_with_id(database):
	doc = dict(foo='bar', _id=1)
	compat.save(database.test_coll, doc)
	assert database.test_coll.find_one() == doc


def test_save_replace_by_id(database):
	compat.save(database.test_coll, dict(foo='bar', _id=1))

	doc = dict(foo='baz', _id=1)
	compat.save(database.test_coll, doc)
	assert database.test_coll.count_documents({}) == 1
	assert database.test_coll.find_one() == doc


def test_save_no_id_extant_docs(database):
	"""
	When no id is supplied, a new document should be created.
	"""
	doc = dict(foo='bar')
	compat.save(database.test_coll, dict(doc))
	assert database.test_coll.count_documents({}) == 1
	compat.save(database.test_coll, doc)
	assert database.test_coll.count_documents({}) == 2


def test_save_adds_id(database):
	"""
	Ensure _id is added to an inserted document.
	"""
	doc = dict(foo='bar')
	compat.save(database.test_coll, doc)
	assert '_id' in doc



import sys
import subprocess
import json


def test_insert_doc_command(mongodb_instance):
	uri = mongodb_instance.get_uri() + '/testdb.test_coll'
	cmd = [
		sys.executable,
		'-m', 'jaraco.mongodb.insert-doc',
		uri,
	]
	proc = subprocess.Popen(cmd, stdin=subprocess.PIPE)
	doc = dict(test='value', test2=2)
	proc.communicate(json.dumps(doc).encode('utf-8'))
	assert not proc.wait()
	saved, = mongodb_instance.get_connection().testdb.test_coll.find()
	saved.pop('_id')
	assert saved == doc



import functools

import pytest

from jaraco.mongodb import testing


@pytest.fixture
def indexed_collection(request, mongodb_instance):
	_name = request.function.__name__
	db = mongodb_instance.get_connection()[_name]
	dropper = functools.partial(db.client.drop_database, _name)
	request.addfinalizer(dropper)
	coll = db[_name]
	coll.create_index('foo')
	return coll


def test_assert_covered_passes(indexed_collection):
	indexed_collection.insert_one({'foo': 'bar'})
	proj = {'_id': False, 'foo': True}
	cur = indexed_collection.find({'foo': 'bar'}, proj)
	testing.assert_covered(cur)


def test_assert_covered_empty(indexed_collection):
	"""
	assert_covered should raise an error it's trivially
	covered (returns no results)
	"""
	cur = indexed_collection.find()
	with pytest.raises(AssertionError):
		testing.assert_covered(cur)


def test_assert_covered_null(indexed_collection):
	"""
	assert_covered should raise an error it's trivially
	covered (returns no results)
	"""
	indexed_collection.insert_one({"foo": "bar"})
	proj = {'_id': False, 'foo': True}
	cur = indexed_collection.find({"foo": "baz"}, proj)
	with pytest.raises(AssertionError):
		testing.assert_covered(cur)


def test_assert_index_used_passes(indexed_collection):
	"""
	assert_index_used should pass when the index is used,
	even if the documents had to be hit.
	"""
	indexed_collection.insert_one({'foo': 'bar', 'bing': 'baz'})
	cur = indexed_collection.find({'foo': 'bar'})
	testing.assert_index_used(cur)


def test_assert_index_used_fails(indexed_collection):
	"""
	assert_index_used should fail when no index is used.
	"""
	indexed_collection.insert_one({'foo': 'bar', 'bing': 'baz'})
	cur = indexed_collection.find({'bing': 'baz'})
	with pytest.raises(AssertionError):
		testing.assert_index_used(cur)



import sys
import json

import argparse

import pymongo.uri_parser


def get_collection(uri):
	parsed = pymongo.uri_parser.parse_uri(uri)
	client = pymongo.MongoClient(uri)
	return client[parsed['database']][parsed['collection']]


def parse_args():
	parser = argparse.ArgumentParser(
		"Insert a document from stdin into the specied collection")
	parser.add_argument(
		'collection',
		metavar='collection_uri',
		type=get_collection,
	)
	return parser.parse_args()


def main():
	args = parse_args()
	args.collection.insert(json.load(sys.stdin))


__name__ == '__main__' and main()



from jaraco.mongodb import fields


def test_insert_with_dots(mongodb_instance):
	db = mongodb_instance.get_connection().test_db
	field = fields.encode("foo.bar")
	db.things.insert_one({field: "value"})
	doc = db.things.find_one({field: "value"})
	doc = {fields.decode(key): value for key, value in doc.items()}
	assert doc['foo.bar'] == 'value'



import functools

import pymongo.collection
from jaraco.collections import Projection


def save(coll, to_save):
	"""
	Pymongo has deprecated the save logic, even though
	MongoDB still advertizes that logic in the core API:
	https://docs.mongodb.com/manual/reference/method/db.collection.save/

	This function provides a compatible interface.
	"""
	filter = Projection(['_id'], to_save)
	upsert_replace = functools.partial(coll.replace_one, filter, upsert=True)
	op = upsert_replace if filter else coll.insert_one
	return op(to_save)


class Collection(pymongo.collection.Collection):
	"""
	Subclass of default Collection that provides a non-deprecated
	save method. Don't use without first reading the cautions at
	https://github.com/mongodb/specifications/blob/master/source/crud/crud.rst#q--a

	>> db = getfixture('database')
	>> coll = Collection(db, 'mycoll')
	>> coll.save({'foo': 'bar'})
	<pymongo.results.InsertOneResult object at ...>
	>> ob = coll.find_one()
	>> ob['foo'] = 'baz'
	>> coll.save(ob)
	<pymongo.results.UpdateResult object at ...>
	"""
	save = save


def query_or_command(op):
	"""
	Given an operation from currentOp, return the query or command
	field as it changed in MongoDB 3.2 for indexing operations:

	https://docs.mongodb.com/manual/reference/method/db.currentOp/#active-indexing-operations
	"""
	return op.get('command') or op.get('query')



from __future__ import print_function

import time
import re
import argparse

from jaraco.mongodb import helper
from .compat import query_or_command


def is_index_op(op):
	cmd = query_or_command(op) or {}
	return 'createIndexes' in cmd


def get_args():
	parser = argparse.ArgumentParser()
	parser.add_argument('db')
	return parser.parse_args()


def run():
	db = helper.connect_db(get_args().db)

	while True:
		ops = db.current_op()['inprog']
		index_op = next(filter(is_index_op, ops), None)
		if not index_op:
			print("No index operations in progress")
			break
		msg = index_op['msg']
		name = query_or_command(index_op)['indexes'][0]['name']
		pat = re.compile(r'Index Build( \(background\))?')
		msg = pat.sub(name, msg, count=1)
		print(msg, end='\r')
		time.sleep(5)


__name__ == '__main__' and run()



import pytest


collect_ignore = [
	'jaraco/mongodb/pmxbot.py',
	# disable move-gridfs check, as it causes output capturing
	# to be disabled. See pytest-dev/pytest#3752.
	'jaraco/mongodb/move-gridfs.py',
]


@pytest.fixture(scope='function')
def database(request, mongodb_instance):
	"""
	Return a clean MongoDB database suitable for testing.
	"""
	db_name = request.node.name.replace('.', '_')
	database = mongodb_instance.get_connection()[db_name]
	yield database
	database.client.drop_database(db_name)
